﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Data.OleDb;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using C1.Win.C1FlexGrid;

namespace C1FlexGridOutlinesAndTrees
{
    public partial class Form1 : Form
    {
        //------------------------------------------------------------------------------------
        #region ** fields

        DataTable _dt = new DataTable();
        CellStyle[] _nodeStyle = new CellStyle[2];

        #endregion

        //------------------------------------------------------------------------------------
        #region ** ctor

        public Form1()
        {
            InitializeComponent();

            // get data
            var fields = @"
                Country, 
                City, 
                SalesPerson, 
                Quantity,
                ExtendedPrice";
            var sql = string.Format("SELECT {0} FROM Invoices ORDER BY {0}", fields);
            var da = new OleDbDataAdapter(sql, GetConnectionString());
            da.Fill(_dt);

            // customize outline styles
            var cs = _nodeStyle[0] = _flex.Styles[CellStyleEnum.Subtotal0];
            cs.BackColor = Color.FromArgb(0xc0, 0xc0, 0xc0);
            cs.ForeColor = Color.Black;
            cs.Font = new Font(_flex.Font, FontStyle.Bold);
            cs = _nodeStyle[1] = _flex.Styles[CellStyleEnum.Subtotal1];
            cs.BackColor = Color.FromArgb(0xe0, 0xe0, 0xe0);
            cs.ForeColor = Color.Black;
            cs.Font = new Font(_flex.Font, FontStyle.Italic);

            // remove vertical gridlines from detail area
            _flex.Styles.Normal.Border.Direction = BorderDirEnum.Horizontal;

            // bind grid to data
            ResetBinding();
        }

        // connection string to c1nwind.mdb (installed with ComponentOne samples)
        static string GetConnectionString()
        {
            string path = Environment.GetFolderPath(Environment.SpecialFolder.Personal) + @"\ComponentOne Samples\Common";
            string conn = @"provider=microsoft.jet.oledb.4.0;data source={0}\c1nwind.mdb;";
            return string.Format(conn, path);
        }

        #endregion

        //------------------------------------------------------------------------------------
        #region ** low-level methods: grouping, trees, subtotals with InsertNode and Aggregate

        // unbind and re-bind grid in order to reset everything
        void ResetBinding()
        {
            // re-bind grid
            _flex.DataSource = null;
            _dt.DefaultView.Sort = string.Empty;
            _flex.Cols.Fixed = fixedColumnToolStripMenuItem.Checked ? 1 : 0;
            _flex.Tree.Column = -1;
            _flex.DataSource = _dt;

            // format ExtendedPrice column
            _flex.Cols["ExtendedPrice"].Format = "n2";

            // resize columns to fit their content
            _flex.AutoSizeCols();

            // allow node content to spill onto next cell
            _flex.AllowMerging = AllowMergingEnum.Nodes;
        }

        // group on a given column inserting nodes of a given level
        void GroupBy(string columnName, int level)
        {
            object current = null;
            for (int r = _flex.Rows.Fixed; r < _flex.Rows.Count; r++)
            {
                if (!_flex.Rows[r].IsNode)
                {
                    var value = _flex[r, columnName];
                    if (!object.Equals(value, current))
                    {
                        // value changed: insert node, apply style
                        _flex.Rows.InsertNode(r, level);
                        _flex.Rows[r].Style = _nodeStyle[Math.Min(level, _nodeStyle.Length - 1)];

                        // show group name in first scrollable column
                        _flex[r, _flex.Cols.Fixed] = value;

                        // update current value
                        current = value;
                    }
                }
            }
        }

        // add subtotals to each node at a given level
        void AddSubtotals(int level, string colName)
        {
            // get column we are going to total on
            int colIndex = _flex.Cols.IndexOf(colName);

            // scan rows looking for nodes at the right level
            for (int r = _flex.Rows.Fixed; r < _flex.Rows.Count; r++)
            {
                if (_flex.Rows[r].IsNode)
                {
                    var node = _flex.Rows[r].Node;
                    if (node.Level == level)
                    {
                        // found a node, calculate the sum of extended price
                        var range = node.GetCellRange();
                        var sum = _flex.Aggregate(AggregateEnum.Sum, 
                            range.r1, colIndex, range.r2, colIndex,
                            AggregateFlags.ExcludeNodes);

                        // show the sum on the grid 
                        // (will use the column format automatically)
                        _flex[r, colIndex] = sum;
                    }
                }
            }
        }

        #endregion

        //------------------------------------------------------------------------------------
        #region ** event handlers

        void _btnReset_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                ResetBinding();
            }
        }
        void _btnGroupCountry_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                ResetBinding();
                GroupBy("Country", 0);
            }
        }
        void _btnGroupCountryCity_Click(object sender, EventArgs e)
        {
            // suspend redrawing while updating
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetBinding();

                // group by Country, City
                GroupBy("Country", 0);
                GroupBy("City", 1);
            }
        }
        void _btnTreeCountry_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // group by country as before
                _btnGroupCountry_Click(this, EventArgs.Empty);

                // show outline tree
                _flex.Tree.Column = 0;

                // autosize to accommodate tree
                _flex.AutoSizeCol(_flex.Tree.Column);

                // collapse detail nodes
                _flex.Tree.Show(0);
            }
        }
        void _btnTreeCountryCity_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // group by country and city as before
                _btnGroupCountryCity_Click(this, EventArgs.Empty);

                // show outline tree
                _flex.Tree.Column = 0;

                // autosize to accommodate tree
                _flex.AutoSizeCol(_flex.Tree.Column);

                // collapse detail nodes
                _flex.Tree.Show(1);
            }
        }
        void _btnNiceTreeCountry_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // start with regular tree
                _btnTreeCountry_Click(this, EventArgs.Empty);

                // add totals per country
                AddSubtotals(0, "ExtendedPrice");  // extended price per country (level 0)
                AddSubtotals(0, "Quantity");       // quantity per country (level 0)

                // hide columns that we grouped on 
                // (they only have duplicate values which already appear on the tree nodes)
                // (but don't make them invisible, that would also hide the node text)
                _flex.Cols["Country"].Width = 0;
            }
        }
        void _btnNiceTreeCountryCity_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // start with regular tree
                _btnTreeCountryCity_Click(this, EventArgs.Empty);

                // add totals per country, city
                AddSubtotals(0, "ExtendedPrice");  // extended price per country (level 0)
                AddSubtotals(0, "Quantity");       // quantity per country (level 0)
                AddSubtotals(1, "ExtendedPrice");  // extended price per city (level 1)
                AddSubtotals(1, "Quantity");       // quantity per city (level 1)

                // hide columns that we grouped on 
                // (they only have duplicate values which already appear on the tree nodes)
                // (but don't make them invisible, that would also hide the node text)
                _flex.Cols["Country"].Width = 0;
                _flex.Cols["City"].Width = 0;
            }
        }

        void _btnGroupCountryHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                ResetBinding();
                _flex.Subtotal(AggregateEnum.None, 0, "Country", "ExtendedPrice");
            }
        }
        void _btnGroupCountryCityHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetBinding();

                // group by Country, City
                _flex.Subtotal(AggregateEnum.None, 0, "Country", "ExtendedPrice");
                _flex.Subtotal(AggregateEnum.None, 1, "City", "ExtendedPrice");
            }
        }
        void _btnTreeCountryHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetBinding();

                // group by country as before
                _flex.Subtotal(AggregateEnum.None, 0, "Country", "ExtendedPrice");

                // show outline tree
                _flex.Tree.Column = 0;
                _flex.AutoSizeCol(_flex.Tree.Column);
                _flex.Tree.Show(0);
            }
        }
        void _btnTreeCountryCityHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetBinding();

                // group by country and city as before
                _flex.Subtotal(AggregateEnum.None, 0, "Country", "ExtendedPrice");
                _flex.Subtotal(AggregateEnum.None, 1, "City", "ExtendedPrice");

                // show outline tree
                _flex.Tree.Column = 0;

                // autosize to accommodate tree
                _flex.AutoSizeCol(_flex.Tree.Column);

                // collapse detail nodes
                _flex.Tree.Show(1);
            }
        }
        void _btnNiceTreeCountryHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetBinding();

                // group and total by country
                _flex.Subtotal(AggregateEnum.Sum, 0, "Country", "ExtendedPrice");
                _flex.Subtotal(AggregateEnum.Sum, 0, "Country", "Quantity");

                // show outline tree
                _flex.Tree.Column = 0;
                _flex.AutoSizeCol(_flex.Tree.Column);
                _flex.Tree.Show(1);

                // hide columns that we grouped on 
                // (they only have duplicate values which already appear on the tree nodes)
                // (but don't make them invisible, that would also hide the node text)
                _flex.Cols["Country"].Width = 0;
            }
        }
        void _btnNiceTreeCountryCityHL_Click(object sender, EventArgs e)
        {
            using (new DeferRefresh(_flex))
            {
                // restore original sort (by Country, City, SalesPerson)
                ResetBinding();

                _flex.Tree.Column = 0;

                // group and total by country and city
                _flex.Subtotal(AggregateEnum.Sum, 0, "Country", "ExtendedPrice");
                _flex.Subtotal(AggregateEnum.Sum, 0, "Country", "Quantity");
                _flex.Subtotal(AggregateEnum.Sum, 1, "City", "ExtendedPrice");
                _flex.Subtotal(AggregateEnum.Sum, 1, "City", "Quantity");

                // hide columns that we grouped on 
                // (they only have duplicate values which already appear on the tree nodes)
                // (but don't make them invisible, that would also hide the node text)
                _flex.Cols["Country"].Width = 0;
                _flex.Cols["City"].Width = 0;
                _flex.AllowMerging = AllowMergingEnum.Nodes;

                // show outline tree
                _flex.Tree.Column = 0;
                _flex.AutoSizeCol(_flex.Tree.Column);
                _flex.Tree.Show(1);
            }
        }

        void fixedColumnToolStripMenuItem_Click(object sender, EventArgs e)
        {
            fixedColumnToolStripMenuItem.Checked = !fixedColumnToolStripMenuItem.Checked;
            ResetBinding();
        }

        #endregion
    }

    /// <summary>
    /// Utility class used to encapsulate grid lengthy operations in a Redraw block.
    /// This avoids flicker and ensures the Redraw property is reset properly in case 
    /// an exception is thrown during the operation.
    /// </summary>
    class DeferRefresh : IDisposable
    {
        C1FlexGrid _grid;
        bool _redraw;
        public DeferRefresh(C1FlexGrid grid)
        {
            _grid = grid;
            _redraw = grid.Redraw;
            grid.Redraw = false;
        }
        public void Dispose()
        {
            _grid.Redraw = _redraw;
        }
    }
}
